/*
    Video: https://www.youtube.com/watch?v=oCMOYS71NIU
    Based on Neil Kolban example for IDF: https://github.com/nkolban/esp32-snippets/blob/master/cpp_utils/tests/BLE%20Tests/SampleNotify.cpp
    Ported to Arduino ESP32 by Evandro Copercini
    updated by chegewara

   Create a BLE server that, once we receive a connection, will send periodic notifications.
   The service advertises itself as: 4fafc201-1fb5-459e-8fcc-c5c9c331914b
   And has a characteristic of: beb5483e-36e1-4688-b7f5-ea07361b26a8

   The design of creating the BLE server is:
   1. Create a BLE Server
   2. Create a BLE Service
   3. Create a BLE Characteristic on the Service
   4. Create a BLE Descriptor on the characteristic
   5. Start the service.
   6. Start advertising.

   A connect hander associated with the server starts a background task that performs notification
   every couple of seconds.
*/

//Libraries 

#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>

//Function headers 

void removeTrailingNulls(String &str); 
void Motor_Control(void);
void Object_Detect_A(void);
void Object_Detect_B(void);
void Battery_Level(void);
void Motor_Current_A(void);
void Motor_Current_B(void); 

//BLE Variables 

BLEServer* pServer = NULL;
BLECharacteristic* pCharacteristic_1 = NULL;
BLECharacteristic* pCharacteristic_2 = NULL; // this defines a pointer for the class BLECharacteristic to get access to all the functions in this class
bool deviceConnected = false;
bool oldDeviceConnected = false;

// ========================================================= Motor Variables ===================================================================

// Motor Variables

//Motor A pins:

int MotorA_ENA = 23;
int MotorA_IN1 = 32;
int MotorA_IN2 = 33; 

//Motor B pins:

int MotorB_ENB = 22;
int MotorB_IN3 = 25;
int MotorB_IN4 = 26;

//Motor Speeds

int MotorA_Speed = 0; 
int MotorB_Speed = 0; 
int xValue_int = 0; 
int yValue_int = 0; 

// ========================================================= Motor Current Monitoring Variables ===================================================================

//Motor A

#define ADC_Pin_Motor_A 39
int adcValue_Motor_A; 
float current_A = 0; 
int senseResistor_A = 3.3; 
const int numSamples_Motor_A = 200;  // Number of samples for averaging
int adcValues_Motor_A[numSamples_Motor_A];  // Array to store ADC values
int totalAdcValue_Motor_A = 0;
float averageAdcValue_Motor_A = 0; 
float averageVoltage_Motor_A = 0; 

//MotoR B

#define ADC_Pin_Motor_B 34
int adcValue_Motor_B = 0; 
float current_B = 0; 
int senseResistor_B = 3.5; 
const int numSamples_Motor_B = 200;  // Number of samples for averaging
int adcValues_Motor_B[numSamples_Motor_B];  // Array to store ADC values
int totalAdcValue_Motor_B = 0;
float averageAdcValue_Motor_B = 0; 
float averageVoltage_Motor_B = 0; 

// ========================================================= Ultrasonic Sensor Variables ===================================================================

#define trigPin_A 16          
#define echoPin_A 17            

#define trigPin_B 14             
#define echoPin_B 27           

float Speed = 0.034; 

long Duration_A = 0; 
float Distance_A = 0; 

long Duration_B = 0; 
float Distance_B = 0; 

// ========================================================= Battery Level Monitoring Variables ===================================================================

//LUT for the Discharge Curve

float Discharge_Curve[] = {
  5.5845401475666972, 2579.999999999999,
  5.8909128631577625, 2519.9999999999955,
  6.093397374693314, 2460.0,
  6.228404238501086, 2400.0,
  6.358547204733114, 2339.9999999999955,
  6.4827800883712854, 2280.0,
  6.603163973091386, 2219.9999999999955,
  6.705782610358694, 2159.9999999999955,
  6.792536671445182, 2100.0,
  6.855266748461503, 2040.0,
  6.9981033717915172, 1980.0000000000005,
  6.940305712466016, 1919.9999999999995,
  6.9771323862940556, 1859.9999999999995,
  7.028364526173304, 1799.9999999999995,
  7.059701750262862, 1739.9999999999995,
  7.094201333189573, 1679.9999999999995,
  7.121929870590815, 1619.9999999999995,
  7.153489034319566, 1560.0,
  7.182395217367992, 1500.0,
  7.216056432882249, 1439.9999999999995,
  7.2556637690575, 1379.9999999999995,
  7.291844156786385, 1319.9999999999995,
  7.325360927051198, 1259.9999999999995,
  7.373381952689952, 1199.9999999999995,
  7.4228331062220282, 1140.0,
  7.466828856999833, 1079.9999999999995,
  7.5270171510629105, 1020.0,
  7.585681057673708, 959.9999999999995,
  7.59730316995594, 899.9999999999995,
  7.642421677421892, 839.9999999999995,
  7.684181629245712, 779.9999999999995,
  7.721978055177991, 719.9999999999995,
  7.7730071921981809, 659.9999999999995,
  7.820250147467505, 599.9999999999995,
  7.8668275856998332, 539.9999999999995,
  7.915681057673708, 479.9999999999998,
  7.959573159973509, 419.9999999999998,
  8.0427974071605292, 359.9999999999998,
  8.084181629245716, 299.9999999999998,
  8.14451604388201, 239.9999999999998,
  8.1915986854092, 179.99999999999989,
  8.205068042455082, 119.99999999999989,
  8.095229793471229, 59.999999999999886,
  8.307978911905941, 0.0
};

//Measured voltage: 
float V_Measured; 
float SoC; 
int Battery_Percentage = 0; 
const int numSamples_Battery = 250;  // Number of samples for averaging
int adcValues_Battery[numSamples_Battery];  // Array to store ADC values
int totalAdcValue_Battery;

//ADC Configuration

#define ADC_PIN 36  // Define the ADC0 pin
#define R1 2660 //top resistor value 
#define R2 970 //bottom resistor value 

// ========================================================= Bluetooth Variables and Callback Functions ===================================================================

// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define SERVICE_UUID        "4fafc201-1fb5-459e-8fcc-c5c9c331914b"
#define CHARACTERISTIC_UUID_1 "beb5483e-36e1-4688-b7f5-ea07361b26a8"
#define CHARACTERISTIC_UUID_2 "194469bc-1efa-4a7d-9e12-eb8ffa43c734"

//Callback Functions 

class MyServerCallbacks: public BLEServerCallbacks { 
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

class CharacteristicsCallBack: public BLECharacteristicCallbacks { //Define callback function that is called when 
    void onWrite(BLECharacteristic* pChar){ //This function is defined using onWrite which allows the client (i.e. phone) to write onto the ESP32 server

      String rxValue = pChar->getValue(); //The getValue() function takes the value (which will be a string) that has been sent from the phone and and stores it in rxValue
      

      if(rxValue.length() > 0){

         removeTrailingNulls(rxValue);

        int separatorIndex = rxValue.indexOf(",");  // Find the position of the comma in the received string

 // If a comma is found in the string
         if (separatorIndex != std::string::npos) { // std::string::npos: This is a constant (-1) representing a value that indicates "no position" or "not found".

        
          // Extract the X value from the beginning of the string to the comma (substr(start, length))
          String xValue = rxValue.substring(0, separatorIndex);
          

          // Extract the Y value from just after the comma to the end of the string
          String yValue = rxValue.substring(separatorIndex + 1);

          // Print the X and Y values to the Serial Monitor
          Serial.print("X: ");
          Serial.print(xValue);
          Serial.print("Y: ");
          Serial.println(yValue);

          //convert String to Integer 

          xValue_int = xValue.toInt();
          yValue_int = yValue.toInt(); 
      }

 //Motor Control: 

    Motor_Control(); 

      }

    }
    
};

//========================================= SETUP =============================================

void setup() {

//========================================= Serial Comms Setup =============================================
  
  Serial.begin(115200);
  Serial.println("Serial communication started.");

//========================================= Ultrasonic Sensor Setup ========================================

  pinMode(trigPin_A, OUTPUT);
  pinMode(echoPin_A, INPUT);

  pinMode(trigPin_B, OUTPUT);
  pinMode(echoPin_B, INPUT);

  //========================================= ADC Setup  ========================================

  analogReadResolution(12); // Set ADC resolution to 12-bit (0-4095)

  //========================================= Motor Setup ==================================================

   //Speed Control Pins: 

    pinMode(MotorA_ENA , OUTPUT);
    pinMode(MotorB_ENB, OUTPUT);

    //Direction Control Pins

    pinMode(MotorA_IN1, OUTPUT);
    pinMode(MotorA_IN2, OUTPUT);
    pinMode(MotorB_IN3, OUTPUT);
    pinMode(MotorB_IN4, OUTPUT);

//========================================= BLE Setup ======================================================

  // Create the BLE Device
  BLEDevice::init("ESP32");

  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pCharacteristic_1 = pService->createCharacteristic( //create new characteristic to recieve from the client and you will need the read and write property
                      CHARACTERISTIC_UUID_1,
                      BLECharacteristic::PROPERTY_READ   |
                      BLECharacteristic::PROPERTY_WRITE  |
                      BLECharacteristic::PROPERTY_NOTIFY |
                      BLECharacteristic::PROPERTY_INDICATE
                    );

   pCharacteristic_2 = pService->createCharacteristic( //create new characteristic to send to the client and you will need the read and write property
                      CHARACTERISTIC_UUID_2,
                       BLECharacteristic::PROPERTY_READ   |
                       BLECharacteristic::PROPERTY_WRITE  |
                       BLECharacteristic::PROPERTY_NOTIFY |
                       BLECharacteristic::PROPERTY_INDICATE
                     );

  // https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.descriptor.gatt.client_characteristic_configuration.xml
  // Create a BLE Descriptor
  BLE2902 *pBLE2902; //declare pointer to an object of type BLE2902 which is a class that controls BLE notifications and indications
  pBLE2902 = new BLE2902(); // dynamically allocate memory for BLE2902 object and the pointer will be directed to this memory address 
  pBLE2902->setNotifications(true); //sets the BLE2902 descriptor to enable notifications.
  
  pCharacteristic_1->addDescriptor(pBLE2902); // adds the BLE2902 descriptor to the characteristic pCharacteristic 
  pCharacteristic_1->setCallbacks(new CharacteristicsCallBack()); // create a new call back function for this characteristic

  pCharacteristic_2->addDescriptor(new BLE2902()); // create descriptor for new characteristic 
 // pCharacteristic_2->setCallbacks(new CharacteristicsCallBack()); // create a new call back function for this characteristic 


  // Start the service
  pService->start();

  // Start advertising
  BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
  pAdvertising->addServiceUUID(SERVICE_UUID);
  pAdvertising->setScanResponse(false);
  pAdvertising->setMinPreferred(0x0);  // set value to 0x00 to not advertise this parameter
  BLEDevice::startAdvertising();
  Serial.println("Waiting a client connection to notify...");

}

  void loop() {
    // notify changed value

     Object_Detect_A();
     Object_Detect_B();
     Battery_Level();
     Motor_Current_A();
     Motor_Current_B();
     //Serial.println("Object detection function called.");

    if (deviceConnected) {

        String dataString = String(int(Distance_A)) + "," + String(int(Distance_B)) + "," + String(Battery_Percentage) + "," + String(int(current_A)) + "," + String(int(current_B)); // convert float to a string
        Serial.print("String length: ");
        Serial.println(dataString.length());
        pCharacteristic_2->setValue(dataString.c_str()); // sets the pCharacteristic value to Object_Distance and converts the C++ String() to C std::char* which is needed for setValue()
        Serial.println(dataString.c_str());
        pCharacteristic_2->notify(); //notifies the connected Bluetooth device that the value of Object Distance was changed  
        delay(250); // bluetooth stack will go into congestion, if too many packets are sent. Delays the next update to avoid that congestion

    }
    // disconnecting
    if (!deviceConnected && oldDeviceConnected) {
        delay(500); // give the bluetooth stack the chance to get things ready
        pServer->startAdvertising(); // restart advertising
        Serial.println("start advertising");
        oldDeviceConnected = deviceConnected;
    }
    // connecting
    if (deviceConnected && !oldDeviceConnected) {
        // do stuff here on connecting
        oldDeviceConnected = deviceConnected;
    }
}

// ========================================= Functions =============================================================

// Function to remove trailing null characters from Arduino String

void removeTrailingNulls(String &str) { // recieve string in dereferenced form so that we can directly modify it within the function 
  if (str.length() > 0 && str.charAt(str.length() - 1) == '\0') { // check if length of string is greater than zero and if the 2nd last character is a null character
    str.remove(str.length() - 1);
  }
}

//Battery level Monitoring Function

void Battery_Level(void){

// Take multiple samples
  for (int i = 0; i < numSamples_Battery; i++) {
    adcValues_Battery[i] = analogRead(ADC_PIN);  // Read the analog value from the ADC2 pin
    totalAdcValue_Battery += adcValues_Battery[i];
    //delay(2.5);  // Short delay between samples
  }

// Calculate the average ADC value
  float averageAdcValue_Battery = totalAdcValue_Battery / numSamples_Battery;
  float V_Measured_Battery = averageAdcValue_Battery * (3.3 / 4095.0)* ((R1 + R2)/R2) + 1.75;  // Convert the average ADC value to voltage
   totalAdcValue_Battery = 0; // reset totalAdcValue

  //Serial.print("Voltage Measured: ");
//  Serial.println(V_Measured);

  if (V_Measured <= 5.5845401475666972){ // Accounts for this special case

      Battery_Percentage = 0; 
     // delay(1000);

    }

  else if (V_Measured >= 8.307978911905941){ //Accounts for this special case 

      Battery_Percentage = 100;
      //delay(1000); 

    }
 
  else {

    for (int i = 0; i < 88; i += 2) { //You go to each voltage value in the array (it is every second value)
  
    if (V_Measured <= Discharge_Curve[i]) { //Check if the measured value is less than the voltage value in the array. When this is true, that means that we have found the range to linearly interpolate
      
      //Can't do the whole calculation in one go so had to break it into chunks
      
      float lower_soc = Discharge_Curve[i - 1];
      float upper_soc = Discharge_Curve[i + 1];
      float lower_voltage = Discharge_Curve[i - 2];
      float upper_voltage = Discharge_Curve[i];

      float soc_slope = (upper_soc - lower_soc) / (upper_voltage - lower_voltage);
      SoC = int(2580 - lower_soc + soc_slope * (V_Measured - lower_voltage)); //this calculate the energy left in the battery 

      Battery_Percentage = int((SoC/2580)*100); //calculate the battery percentage 
      
      //Serial.print("SoC: ");
      //Serial.println(SoC);
      //Serial.print("Battery Percentage: ");
      //Serial.println(Battery_Percentage);
      //delay(100);
     // break;
    }
  }
 }

}

//Motor Current Monitoring 

void Motor_Current_A(void){

    // Take multiple samples
    for (int i = 0; i < numSamples_Motor_A; i++) {
      adcValues_Motor_A[i] = analogRead(ADC_Pin_Motor_A);  // Read the analog value from the ADC pin
      totalAdcValue_Motor_A += adcValues_Motor_A[i];
      //delay(5);  // Short delay between samples
    }

    // Calculate the average ADC value
    averageAdcValue_Motor_A = totalAdcValue_Motor_A / (float)numSamples_Motor_A;
    averageVoltage_Motor_A = averageAdcValue_Motor_A * (3.3 / 4095.0) + 0.135;  // Convert the average ADC value to voltage

    if(averageAdcValue_Motor_A == 0){ //this is to negate the effects of the calibration value 
        averageVoltage_Motor_A = 0; 
    }

    current_A = (averageVoltage_Motor_A / 3.3) * 1000;  // Calculate the current

    totalAdcValue_Motor_A = 0; // Reset

    // Print the results
    Serial.print("Average ADC Value (Motor A): ");
    Serial.print(averageAdcValue_Motor_A);  
    Serial.print("Average Voltage (Motor A): ");
    Serial.print(averageVoltage_Motor_A, 2);  
    Serial.print("Current (Motor A): ");
    Serial.println(current_A);  
}

void Motor_Current_B(void){

    // Take multiple samples
    for (int i = 0; i < numSamples_Motor_B; i++) {
      adcValues_Motor_B[i] = analogRead(ADC_Pin_Motor_B);  // Read the analog value from the ADC pin
      totalAdcValue_Motor_B += adcValues_Motor_B[i];
      //delay(5);  // Short delay between samples
    }

    // Calculate the average ADC value
    averageAdcValue_Motor_B = totalAdcValue_Motor_B / (float)numSamples_Motor_B;
    averageVoltage_Motor_B = averageAdcValue_Motor_B * (3.3 / 4095.0) + 0.135;  // Convert the average ADC value to voltage

    if(averageAdcValue_Motor_B == 0){ //this is to negate the effects of the calibration value 
        averageVoltage_Motor_B = 0; 
    }

    current_B = (averageVoltage_Motor_B / 3.8) * 1000;  // Calculate the current

    totalAdcValue_Motor_B = 0; // Reset

    // Print the results
    Serial.print("Average ADC Value (Motor B): ");
    Serial.print(averageAdcValue_Motor_B);  
    Serial.print("Average Voltage (Motor B): ");
    Serial.print(averageVoltage_Motor_B, 2);  
    Serial.print("Current (Motor B): ");
    Serial.println(current_B);  
}

// Object Deterction Function 

void Object_Detect_A(){

//sends a 10us pulse to trigger the transmission of the ultrasonic burst and the ECHO pin is automatically made HIGH.

  digitalWrite(trigPin_A, LOW);
  delayMicroseconds(2);
  digitalWrite(trigPin_A, HIGH); 
  delayMicroseconds(10);
  digitalWrite(trigPin_A, LOW); 

//Note: The ECHO pin goes low after the ultrasonic burst returns
  
  Duration_A = pulseIn(echoPin_A, HIGH); // the pulseIn() function determines how long the ECHO pin remains high for to determine the time taken for the Ultrasonic burst to leave and return to the sensor. 
  Distance_A = (Duration_A*Speed)/2;

  // Serial.print("Distance: "); 
  // Serial.println(Distance_B); 
  // Serial.println(" cm"); // print “cm” with a newline character so that the next value is on the next line 

 //delay(25); //Delay before the characteristic is updated

}

void Object_Detect_B(){

//sends a 10us pulse to trigger the transmission of the ultrasonic burst and the ECHO pin is automatically made HIGH.

  digitalWrite(trigPin_B, LOW);
  delayMicroseconds(2);
  digitalWrite(trigPin_B, HIGH); 
  delayMicroseconds(10);
  digitalWrite(trigPin_B, LOW); 

//Note: The ECHO pin goes low after the ultrasonic burst returns

  Duration_B = pulseIn(echoPin_B, HIGH); // the pulseIn() function determines how long the ECHO pin remains high for to determine the time taken for the Ultrasonic burst to leave and return to the sensor. 
  Distance_B = (Duration_B*Speed)/2;

  // Serial.print("Distance: "); 
  // Serial.println(Distance_B); 
  // Serial.println(" cm"); // print “cm” with a newline character so that the next value is on the next line 

 //delay(25); //Delay before the characteristic is updated

}

// Motor Control Function 

void Motor_Control(){

// Y- Axis used for forward and backward movement 

if (yValue_int > 100) {

//Set Motor A backwards
digitalWrite(MotorA_IN1, HIGH);
digitalWrite(MotorA_IN2, LOW);

//Set Motor B backwards
digitalWrite(MotorB_IN3, HIGH);
digitalWrite(MotorB_IN4, LOW); 

//Convert Position to speed (map the position of the joystick to the required speed constrained between 3V - 6V)
MotorA_Speed = map(yValue_int, 100, 200, 120, 255);
MotorB_Speed = map(yValue_int, 100, 200, 120, 255);
Serial.print("Motor Speed B: ");
Serial.print(MotorB_Speed); 
Serial.print("Motor Speed A: ");
Serial.println(MotorA_Speed);


}
else if (yValue_int < 100) {

//Set Motor A forwards
digitalWrite(MotorA_IN1, LOW);
digitalWrite(MotorA_IN2, HIGH);

//Set Motor B backwards
digitalWrite(MotorB_IN3, LOW);
digitalWrite(MotorB_IN4, HIGH); 

//Convert Position to speed (map the position of the joystick to the required speed constrained between 3V - 6V)
MotorA_Speed = map(yValue_int, 100, 0, 120, 255);
MotorB_Speed = map(yValue_int, 100, 0, 120, 255);

Serial.print("Motor Speed B: ");
Serial.print(MotorB_Speed); 
Serial.print("Motor Speed A: ");
Serial.println(MotorA_Speed);

}

// X - Axis used for left and right movement (this is occurs when both motors move at different speeds)

if (xValue_int < 100) {

//Convert Position to speed (map the position of the joystick to the required speed constrained between 3V - 6V)
int xMapped = map(xValue_int, 100, 0, 120, 255);

Serial.print("xMapped: "); 
Serial.println(xMapped);

//Move to the left - achieved by increasing speed of right motor and decreasing speed of left motor 

MotorA_Speed = MotorA_Speed + xMapped/7; 
MotorB_Speed = MotorB_Speed - xMapped/7; 

// Confine the range from 90 - 170 because there are positions that of the joystick that results in MotorA_Speed < 90 and MotorB_Speed > 170

if (MotorA_Speed > 255) {

    MotorA_Speed = 255; 

}

if (MotorB_Speed < 120) {

    MotorB_Speed = 120; 

}

}

else if (xValue_int > 100) {

//Convert Position to speed (map the position of the joystick to the required speed constrained between 3V - 6V)
int xMapped = map(xValue_int, 100, 200, 120, 255);

Serial.print("xMapped: "); 
Serial.println(xMapped);

//Move to the right - achieved by increasing speed of right motor and decreasing speed of left motor

MotorA_Speed = MotorA_Speed - xMapped/7; 
MotorB_Speed = MotorB_Speed + xMapped/7; 

// Confine the range from 90 - 170 because there are positions that of the joystick that results in MotorA_Speed > 170 and MotorB_Speed < 90

if (MotorA_Speed < 120){

    MotorA_Speed = 120;

}

if (MotorB_Speed > 255){

    MotorB_Speed = 255;

}

}

analogWrite(MotorA_ENA, MotorA_Speed);
analogWrite(MotorB_ENB, MotorB_Speed); 

Serial.print("Motor Speed B: ");
Serial.print(MotorB_Speed); 
Serial.print("Motor Speed A: ");
Serial.println(MotorA_Speed);


if (xValue_int ==  100 && yValue_int == 100){ //Switch off motors when joystick goes back to the middle 

//Stop Motor A 
digitalWrite(MotorA_IN1, LOW);
digitalWrite(MotorA_IN2, LOW);

//Stop Motor B 
digitalWrite(MotorB_IN3, LOW);
digitalWrite(MotorB_IN4, LOW);

}

}